#include "XC4607.h"

#define FRAME_TIME 200

#define LEFT_BUTTON A5
#define MIDDLE_BUTTON A3
#define RIGHT_BUTTON A4

int paddle_pins[] = {A0, A1};
int player_button_pins[] = {LEFT_BUTTON, RIGHT_BUTTON};

enum GameState {
  START, SELECT, SERVE, PLAY, WIN
};

unsigned int start_image[] = {
  0b0111000000000001,
  0b1000100000000001,
  0b1000100000000001,
  0b1111101011001111,
  0b1000101100010001,
  0b1000101000010001,
  0b1000101000001111,
  0b0000000000000000,
  0b1111000000000000,
  0b1000100000000000,
  0b1111000000000111,
  0b1000000000001001,
  0b1001100111001001,
  0b1010010100100111,
  0b1010010100100001,
  0b1001100100101110,
};

unsigned int win_image[] = {
  0b10001011101001,
  0b10101001001101,
  0b10101001001011,
  0b01010011101001,
};

byte letter_font_data[] = {
  0b000, 0b000, 0b010, 0b011, 0b111, 0b011, 0b101, 0b101, 0b101, 0b110, 0b110, 0b011, 0b111, 0b101,
  0b010, 0b010, 0b101, 0b100, 0b100, 0b100, 0b101, 0b111, 0b101, 0b101, 0b101, 0b100, 0b010, 0b101,
  0b110, 0b011, 0b111, 0b100, 0b111, 0b101, 0b111, 0b101, 0b101, 0b110, 0b110, 0b010, 0b010, 0b010,
  0b010, 0b010, 0b101, 0b100, 0b100, 0b101, 0b101, 0b101, 0b101, 0b100, 0b101, 0b001, 0b010, 0b101,
  0b000, 0b000, 0b101, 0b011, 0b111, 0b011, 0b101, 0b101, 0b101, 0b100, 0b101, 0b110, 0b010, 0b101,
};

Font letter_font = {
  3, 5, 14,
  "<>ACEGHMOPRSTX",
  letter_font_data
};

byte score_font_data[] = {
  0b010, 0b010, 0b110, 0b110, 0b001, 0b111, 0b011, 0b111, 0b010, 0b010,
  0b101, 0b110, 0b001, 0b001, 0b011, 0b100, 0b100, 0b001, 0b101, 0b101,
  0b101, 0b010, 0b110, 0b110, 0b101, 0b110, 0b110, 0b010, 0b010, 0b011,
  0b101, 0b010, 0b100, 0b001, 0b111, 0b001, 0b101, 0b010, 0b101, 0b001,
  0b010, 0b111, 0b111, 0b110, 0b001, 0b110, 0b010, 0b010, 0b010, 0b110,
};

Font score_font = {
  3, 5, 10,
  "0123456789",
  score_font_data
};

byte player_font_data[] = {
  0b00100, 0b11110,
  0b01010, 0b10001,
  0b10001, 0b11110,
  0b11111, 0b10001,
  0b10001, 0b11110,
};

Font player_font = {
  5, 5, 2,
  "AB",
  player_font_data
};

char *player_letters[] = {"A", "B"};

XC4607 disp;
int ballx, bally;
int ballvx, ballvy;
int ball_slowness = 4;
int ball_move_counter;
int paddle[2];
int paddle_size = 5;
int serving_player;
int scores[2];
int winning_player;
int max_score = 9;
byte odd_frame;
GameState state;

void init_timer_interrupt() {
  // Interrupt when Timer 0 reaches 0xAF
  OCR0A = 0xAF;
  TIMSK0 |= _BV(OCIE0A);
}

// Millisecond interrupt ISR
ISR(TIMER0_COMPA_vect) {
  disp.tick();
//  digitalWrite(LED_BUILTIN, digitalRead(A5));
}

bool button(int pin) {
  return !digitalRead(pin);
}

bool player_button(int player) {
  return button(player_button_pins[player]);
}


void update_paddles() {
  for (int i = 0; i <= 1; i++) {
    int a = constrain(analogRead(paddle_pins[i]) - 24, 0, 1000 - 24);
    int p = a * (16 - paddle_size) / (1000 - 24);
    paddle[i] = p;
  }
}

int paddle_top(int i) {
  return paddle[i];
}

int paddle_middle(int i) {
  return paddle[i] + (paddle_size >> 1);
}

int paddle_bottom(int i) {
  return paddle[i] + paddle_size - 1;
}

bool paddle_hit(int i) {
  if (bally >= paddle_top(i) && bally <= paddle_bottom(i)) {
    if (i == 0)
      return ballx == 1;
    else
      return ballx == 14;
  }
  else
    return false;
}

void move_ball() {
  if (--ball_move_counter <= 0) {
    ball_move_counter = ball_slowness;
    if ((ballvx < 0 && paddle_hit(0)) || (ballvx > 0 && paddle_hit(1)))
      ballvx = -ballvx;
    if ((ballvy < 0 && bally == 0) || (ballvy > 0 && bally == 15))
      ballvy = -ballvy;
    ballx = ballx + ballvx;
    bally = bally + ballvy;
  }
}

void move_ball_to_serving_position(int player) {
  switch (player) {
    case 0:
      ballx = 1;
      break;
    case 1:
      ballx = 14;
      break;
  }
  bally = paddle_middle(player);
}

int draw_paddles() {
  disp.vline(0, paddle_top(0), paddle_size);
  disp.vline(15, paddle_top(1), paddle_size);
}

void draw_ball() {
  disp.set(ballx, bally);
}

void draw_score(int score, int x, int y) {
  char buf[2];
  sprintf(buf, "%d", score);
  disp.write(x, y, &score_font, buf);
}

void draw_scores() {
  draw_score(scores[0], 4, 0);
  draw_score(scores[1], 9, 0);
}

void serve(int player) {
  switch(player) {
    case 0:
      ballvx = 1;
      break;
    case 1:
      ballvx = -1;
      break;
  }
  ballvy = 1;
  state = PLAY;
}

int ball_out_of_bounds() {
  if (ballx < 0)
    return 0;
  else if (ballx > 15)
    return 1;
  else
    return -1;
}

void new_game() {
  state = SERVE;
  scores[0] = 0;
  scores[1] = 0;
}

void score_point(int player) {
  scores[player] += 1;
  if (scores[player] == max_score) {
    state = WIN;
    winning_player = player;
  }
  else {
    state = SERVE;
    serving_player = player;
  }
}

void start_or_select() {
  if (button(LEFT_BUTTON) || button(RIGHT_BUTTON))
    state = SELECT;
  else if (button(MIDDLE_BUTTON))
    new_game();
}

void setup() {
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(LEFT_BUTTON, INPUT_PULLUP);
  pinMode(MIDDLE_BUTTON, INPUT_PULLUP);
  pinMode(RIGHT_BUTTON, INPUT_PULLUP);
  init_timer_interrupt();
  state = START;
  serving_player = 0;
}

void loop() {
  odd_frame ^= 1;
  disp.wait_vsync();
  disp.cls();
  switch (state) {
    case START: {
      disp.blit16(0, 0, start_image, 16, 16);
      start_or_select();
      break;
    }
    case SELECT: {
      draw_score(max_score, 2, 0);
      disp.write(7, 0, &letter_font, "PT");
      disp.write(0, 6, &letter_font, "GAME");
      disp.write(0, 11, &letter_font, "<");
      disp.write(13, 11, &letter_font, ">");
      if (button(LEFT_BUTTON))
        max_score = 3;
      else if (button(RIGHT_BUTTON))
        max_score = 9;
      else if (button(MIDDLE_BUTTON))
        new_game();
      break;
    }
    case SERVE: {
      update_paddles();
      move_ball_to_serving_position(serving_player);
      draw_scores();
      draw_paddles();
      draw_ball();
      if (player_button(serving_player))
        serve(serving_player);
      break;
    }
    case PLAY: {
      update_paddles();
      move_ball();
      draw_scores();
      draw_paddles();
      int missing_player = ball_out_of_bounds();
      if (missing_player >= 0)
        score_point(1 - missing_player);
      else
        draw_ball();
      break;
    }
    case WIN: {
      draw_scores();
      disp.write(6, 6, &player_font, player_letters[winning_player]);
      disp.blit16(1, 12, win_image, 14, 4);
      start_or_select();
      break;
    }
  }
}
